/*
 * Copyright (C) 2002  Olaf Beck <olaf_sc@yahoo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#define MAJOR_VER 0
#define MINOR_VER 2
#define SUB_VER 1

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef WIN32
#define uint64_t unsigned __int64
#define int64_t __int64
#define uint32_t unsigned  int
#define int32_t int;
#define uint16_t unsigned short
#define int16_t short
#define uint8_t unsigned char
#define int8_t char
#include <memory.h>
#include <io.h>
#include "getopt.h"

/* Need to ckeck if this is really needed */
#define read _read
#define write _write
#define open _open
#define lseek _lseek

#else  /* Unix */
#include <inttypes.h>
#include <unistd.h>
#endif



#if defined(__MINGW32__)
#include <string.h>
#endif


#ifdef A52
#include "a52.h"
#include "audio_out.h"
#include "mm_accel.h"

#include <string.h>
#include <math.h>

#include "avcodec.h"
#endif



/**
 * Storage for constructed mpeg parts
 */

/**
 * The Audio PES - linked list so we
 * can automatically sort the entries
 * pes holds a whole pes pack of
 * either AC3 or MPEG audio
 */

typedef struct Audio_Pes_t audio_pes_t;

struct Audio_Pes_t {
	uint64_t time;
	uint32_t size;
	int pes_header_record_size;
	int pes_header_record_offset;
	uint8_t * pes_header;
	uint8_t * pes_data;
	audio_pes_t * next;
	audio_pes_t * previous;
};


/**
 * The Video PES - linked list so we
 * can automatically sort the entries
 * pes_data holds a whole pes payload of
 * MPEG video
 */

typedef struct Video_Pes_t video_pes_t;

struct Video_Pes_t {
	uint64_t time;
	uint32_t size;
	uint8_t * pes_header;
	uint8_t * pes_data;
	video_pes_t * next;
	video_pes_t * previous;
};


/**
 * Tivo structures
 */

/**
 * Tivo has 128KB chunks that stores
 * the MPEG data
 * The structure is as below:
 * - chunk_header
 * - records * record_header (16 byte record header)
 * - data
 */

/**
 * Chunk header not used just info
 */

/*
typedef struct {
	uint16_t nr_records;
	uint16_t unknown;
} ATTRIBUTE_PACKED chunk_header_t;
*/

/**
 * Record header
 */

typedef struct {
	uint32_t  size;
	uint16_t  type;
	uint32_t  junk1;
	uint32_t  junk2;
	uint32_t  junk3;
} record_header_t;


/**
 * Records
 */

typedef struct {
	uint8_t * record;
} record_t;


/**
 * Storage area for the Tivo
 * chunk while assembling the
 * output MPEG data
 */

typedef struct Chunk_t chunk_t;

struct Chunk_t {
	uint16_t nr_records;
	uint16_t unknown;
	record_header_t * record_header;
	record_t * records;

	/* Access pointers */
	chunk_t * next ;
	chunk_t * previous;
};




/* END Global variables */


/* Prototypes */

/**
 * Chunk operations
 */

/**
 * Prints the record size, type and a max of 16 bytes of the record
 * data of all present records in a chunk
 */

void print_chunk(chunk_t * chunk);

/**
 * Free's all data allocated in a chunk
 */

void free_chunk(chunk_t * chunk);

/**
 * Free's all data allocated in the chunk
 * linked list
 */

void free_chunks(chunk_t * chunks);

/**
 * Insets a chunk to the chunk linked list and returns
 * a pointer to the first chunk in the linked list.
 * chunk == the chunk to add
 * chunks == the chunk linked list
 *
 * The function does a check if the chunk added is a
 * duplication of the previously added chunk.
 */

chunk_t * add_chunk(chunk_t * chunk, chunk_t * chunks);

/**
 * Parses the first chunk in the chunk linked list
 * and inserts audio and video pes data
 * to the video and audio pes linked list.
 * Returns the next chunk in the chunk linked list.
 */

chunk_t * parse_chunk(chunk_t * chunks);

/**
 * Reads the record header in a chunk present in the file_nr
 * Retunes a record_header structurs.
 */
record_header_t * read_record(int nr_records, int file_nr);

/**
 * Reads the chunk present in the file_nr. Retuns a
 * chunk structue. The functio has error detection and
 * will return NULL if the chunk is bad. NOTE: You need
 * to do a lseek to the right location before calling this
 * function
 */
chunk_t * read_chunk(int file_nr);


/**
 * Check if the chunk is good or bad. 
 * 
 * Checks:
 * If the audio in the chunk is off correct type
 * If there is junk record types
 * If the video is out of sync with previus chunk
 *
 * If any of the above fails it will return NULL,
 * other wise it will return the checked chunk.
 */
chunk_t * check_chunk(chunk_t * chunk);



/********************************************/

/**
 * Video PES functions
 */

/**
 * Free's all data allocated in a video pes
 */

void free_video_pes(video_pes_t * video_pes);

/**
 * Free's all pes present in the video pes linked list
 */

void free_video_pess(video_pes_t * video_pes);

/**
 * Writes a video pes to a file, and returns a pointer
 * to the next pes in the pes linked list or null
 * if it was the last pes in the list.
 */

video_pes_t * write_video_pes(video_pes_t * video_pes);


/**
 * Inserts a video pes to the end of the pes linked list
 * and returns a pointer to the first pes in the pes linked
 * linked list.
 */

video_pes_t * insert_video_pes(video_pes_t * video_pes);

/**
 * Returns the last pes in the video pes linked list
 */

video_pes_t * get_last_video_pes();

/*********************************************/

/**
 * Audio Pes functions
 */

/**
 * Free's all data allocated in a audio pes
 */

void free_audio_pes(audio_pes_t * audio_pes);

/**
 * Free's all pes present in the audio pes linked list
 */

void free_audio_pess(audio_pes_t * audio_pes);

/**
 * Writes a audio pes to a file, and returns a pointer
 * to the next pes in the pes linked list or null
 * if it was the last pes in the list.
 */

audio_pes_t * write_audio_pes(audio_pes_t * audio_pes);

/**
 * Inserts a video pes to the end of the pes linked list
 * and returns a pointer to the first pes in the pes linked
 * linked list.
 */

audio_pes_t * insert_audio_pes(audio_pes_t * audio_pes);

/**
 * Returns the last pes in the audio pes linked list
 */

audio_pes_t * get_last_audio_pes();


/**
 * Test function will be removed or rewritten don't use it!
 */

audio_pes_t * print_it(audio_pes_t * audio_pes);



/** 
 * Video / Audio parsing 
 */
 
void parse_video(chunk_t * chunks);
void parse_audio(chunk_t * chunks);




/**
 * Misc functions
 */

/**
 * Returns the time in ticks, the uint8_t pointer
 * must point to the begining of the PTS or DTS in
 * the pes header i.e. pt = pes_header + 9.
 */

uint64_t get_time(uint8_t * pt);

/**
 * Returns the type of the tystream (although all vital info is
 * set in the function) 1 == AC3 2 == MPEG Layer II. 
 * Returns 0 on failur
 */
int probe_tystream(int file_nr, off_t file_size);

/** 
 * Seeks the start of a right_audio type in the tystream, 0x9c0 == AC3,
 * 0x3c0 == MPEG Layer II - sets the file offset so you can start reading 
 * the chunk right after this function
 */

void get_start_chunk(int file_nr, uint16_t right_audio, off_t file_size);

/**
 * Prints the progress - FIXME need a better one time, % etc.. 
 */

void prog_meter(int chunk_lnr);

/**
 * Prints usage and exits the prog with 1
 */

void usage();
